from __future__ import annotations

from typing import Any, Iterable, Iterator, Literal, Sequence, TypeVar, overload

from ..delayed import Delayed

T_co = TypeVar("T_co", covariant=True)

class Array(Iterable[T_co]):
    """Distributed array-like collection."""

    def compute(
        self,
        *,
        scheduler: str | None = ...,
        optimize_graph: bool = ...,
        traverse: bool | None = ...,
        **kwargs: Any,
    ) -> T_co: ...
    def persist(
        self,
        *,
        scheduler: str | None = ...,
        optimize_graph: bool = ...,
        traverse: bool | None = ...,
        **kwargs: Any,
    ) -> Array[T_co]: ...
    def to_delayed(self) -> list[list[Delayed[T_co]]]: ...
    def __iter__(self) -> Iterator[T_co]: ...

_ShapeLike = Sequence[int] | tuple[int, ...]
_ChunksLike = Sequence[int] | tuple[int, ...] | Literal["auto"] | None

@overload
def from_delayed(
    value: Delayed[T_co],
    shape: _ShapeLike,
    dtype: Any,
    *,
    meta: Any = ...,
    name: str | None = ...,
    verify_meta: bool = ...,
    nout: None = ...,
    chunks: _ChunksLike = ...,
) -> Array[T_co]: ...
@overload
def from_delayed(
    values: Sequence[Delayed[T_co]],
    shape: _ShapeLike,
    dtype: Any,
    *,
    meta: Any = ...,
    name: str | None = ...,
    verify_meta: bool = ...,
    nout: int,
    chunks: _ChunksLike = ...,
) -> Array[T_co]: ...

__all__ = ["Array", "from_delayed"]



from __future__ import annotations

from collections.abc import Awaitable
from typing import Any, Protocol

class _CommClosedError(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> Exception: ...

class Scheduler:
    """Coordinate work across distributed workers."""

    address: str
    loop: Any
    comm_closed_handler: _CommClosedError | None

    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    async def close(self, fast: bool | None = ...) -> None: ...
    async def finished(self) -> None: ...
    async def start(self, *args: Any, **kwargs: Any) -> None: ...
    def start_http_server(self, *args: Any, **kwargs: Any) -> None: ...
    def add_plugin(self, plugin: Any, name: str | None = ...) -> None: ...
    def call_client_handlers(self, key: str, **kwargs: Any) -> Awaitable[Any]: ...
    def set_comm_closed_handler(self, handler: _CommClosedError | None) -> None: ...

__all__ = ["Scheduler"]



from __future__ import annotations

from collections.abc import Iterator, MutableMapping, Sequence
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    Literal,
    Mapping,
    ParamSpec,
    TypeVar,
    overload,
)

from dask.delayed import Delayed
from .scheduler import Scheduler
from .worker import Worker, WorkerPlugin

T = TypeVar("T")
U = TypeVar("U")
KT = TypeVar("KT")
P = ParamSpec("P")
T_co = TypeVar("T_co", covariant=True)
PersistableT = TypeVar("PersistableT")

class Future(Generic[T_co]):
    """Remote result managed by a Dask scheduler."""

    key: str
    status: str
    client: Client

    def cancel(self) -> bool: ...
    def cancelled(self) -> bool: ...
    def done(self) -> bool: ...
    def result(self, timeout: float | None = ..., **kwargs: Any) -> T_co: ...
    def exception(self, timeout: float | None = ...) -> BaseException | None: ...
    def add_done_callback(self, fn: Callable[[Future[T_co]], Any]) -> None: ...
    def release(self) -> None: ...
    def to_delayed(self) -> Delayed[T_co]: ...
    def __await__(self) -> Generator[Any, Any, T_co]: ...
    def __repr__(self) -> str: ...

class AsCompleted(Iterable[Future[Any]]):
    """Iterator that yields futures as they finish."""

    def __iter__(self) -> Iterator[Future[Any]]: ...
    def results(self, timeout: float | None = ...) -> list[Any]: ...

class Client:
    """Synchronous entry point to a Dask scheduler."""

    asynchronous: bool
    scheduler: str | None
    planner: str | None

    def __init__(
        self,
        address: str | None = ...,
        *,
        loop: Any = ...,
        asynchronous: bool | None = ...,
        timeout: float | None = ...,
        set_as_default: bool = ...,
        **kwargs: Any,
    ) -> None: ...
    def close(self, *, timeout: float | None = ...) -> None: ...
    def shutdown(self, *, timeout: float | None = ...) -> None: ...
    def restart(self, timeout: float | None = ...) -> None: ...
    def submit(self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs) -> Future[T]: ...
    def map(
        self,
        func: Callable[..., T],
        *iterables: Iterable[Any],
        key: str | None = ...,
        **kwargs: Any,
    ) -> list[Future[T]]: ...
    @overload
    @overload
    def gather(
        self,
        futures: Future[T],
        *,
        errors: str = ...,
        direct: bool = ...,
        asynchronous: Literal[True],
    ) -> Awaitable[T]: ...
    @overload
    def gather(
        self,
        futures: Sequence[Future[T]],
        *,
        errors: str = ...,
        direct: bool = ...,
        asynchronous: Literal[True],
    ) -> Awaitable[list[T]]: ...
    @overload
    def gather(
        self,
        futures: Future[T],
        *,
        errors: str = ...,
        direct: bool = ...,
        asynchronous: Literal[False] | None = ...,
    ) -> T: ...
    @overload
    def gather(
        self,
        futures: Sequence[Future[T]],
        *,
        errors: str = ...,
        direct: bool = ...,
        asynchronous: Literal[False] | None = ...,
    ) -> list[T]: ...
    @overload
    def scatter(
        self,
        data: Mapping[KT, T],
        *,
        broadcast: bool = ...,
        direct: bool = ...,
        hash: bool | None = ...,
        asynchronous: Literal[True],
    ) -> Awaitable[MutableMapping[KT, Future[T]]]: ...
    @overload
    def scatter(
        self,
        data: Mapping[KT, T],
        *,
        broadcast: bool = ...,
        direct: bool = ...,
        hash: bool | None = ...,
        asynchronous: Literal[False] | None = ...,
    ) -> MutableMapping[KT, Future[T]]: ...
    @overload
    def scatter(
        self,
        data: Sequence[T],
        *,
        broadcast: bool = ...,
        direct: bool = ...,
        hash: bool | None = ...,
        asynchronous: Literal[True],
    ) -> Awaitable[list[Future[T]]]: ...
    @overload
    def scatter(
        self,
        data: Sequence[T],
        *,
        broadcast: bool = ...,
        direct: bool = ...,
        hash: bool | None = ...,
        asynchronous: Literal[False] | None = ...,
    ) -> list[Future[T]]: ...
    @overload
    def compute(
        self,
        collection: Delayed[T],
        *,
        optimize_graph: bool = ...,
        scheduler: str | None = ...,
        asynchronous: Literal[True],
    ) -> Awaitable[T]: ...
    @overload
    def compute(
        self,
        collection: Any,
        *,
        optimize_graph: bool = ...,
        scheduler: str | None = ...,
        asynchronous: Literal[True],
    ) -> Awaitable[Any]: ...
    @overload
    def compute(
        self,
        collection: Delayed[T],
        *,
        optimize_graph: bool = ...,
        scheduler: str | None = ...,
        asynchronous: Literal[False] | None = ...,
    ) -> T: ...
    @overload
    def compute(
        self,
        collection: Any,
        *,
        optimize_graph: bool = ...,
        scheduler: str | None = ...,
        asynchronous: Literal[False] | None = ...,
    ) -> Any: ...
    @overload
    def persist(
        self,
        collection: PersistableT,
        *,
        optimize_graph: bool = ...,
        scheduler: str | None = ...,
        traverse: bool | None = ...,
        asynchronous: Literal[True],
    ) -> Awaitable[PersistableT]: ...
    @overload
    def persist(
        self,
        collection: PersistableT,
        *,
        optimize_graph: bool = ...,
        scheduler: str | None = ...,
        traverse: bool | None = ...,
        asynchronous: Literal[False] | None = ...,
    ) -> PersistableT: ...
    def get_future(self, key: str) -> Future[Any]: ...
    def cancel(self, futures: Future[Any] | Sequence[Future[Any]]) -> None: ...
    def futures_of(self, collection: Any) -> list[Future[Any]]: ...
    def publish_dataset(self, name: str, data: Any, *, override: bool = ...) -> None: ...
    def list_datasets(self) -> Dict[str, str]: ...
    def get_dataset(self, name: str) -> Any: ...
    def run(self, function: Callable[..., U], *args: Any, **kwargs: Any) -> Dict[str, U]: ...
    def run_on_scheduler(self, function: Callable[..., U], *args: Any, **kwargs: Any) -> U: ...
    def wait_for_workers(self, n_workers: int = ..., timeout: float | None = ...) -> None: ...
    def who_has(
        self,
        futures: Future[Any] | Sequence[Future[Any]] | None = ...,
        **kwargs: Any,
    ) -> Mapping[str, list[str]]: ...
    def has_what(
        self,
        workers: str | Sequence[str] | None = ...,
    ) -> Mapping[str, list[str]]: ...
    def sync(
        self,
        func: Callable[..., U],
        *args: Any,
        asynchronous: bool | None = ...,
        **kwargs: Any,
    ) -> U: ...
    def scheduler_info(self) -> Dict[str, Any]: ...
    def __await__(self) -> Generator[Any, Any, AsyncClient]: ...
    def __enter__(self) -> Client: ...
    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None: ...
    async def __aenter__(self) -> AsyncClient: ...
    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None: ...

class AsyncClient:
    """Async wrapper around :class:`Client`."""

    asynchronous: bool

    def __init__(
        self,
        address: str | None = ...,
        *,
        loop: Any = ...,
        timeout: float | None = ...,
        set_as_default: bool = ...,
        **kwargs: Any,
    ) -> None: ...
    async def submit(
        self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> Future[T]: ...
    async def map(
        self,
        func: Callable[..., T],
        *iterables: Iterable[Any],
        key: str | None = ...,
        **kwargs: Any,
    ) -> list[Future[T]]: ...
    async def close(self, *, timeout: float | None = ...) -> None: ...
    async def shutdown(self, *, timeout: float | None = ...) -> None: ...
    @overload
    async def gather(
        self,
        futures: Future[T],
        *,
        errors: str = ...,
        direct: bool = ...,
    ) -> T: ...
    @overload
    async def gather(
        self,
        futures: Sequence[Future[T]],
        *,
        errors: str = ...,
        direct: bool = ...,
    ) -> list[T]: ...
    async def who_has(
        self,
        futures: Future[Any] | Sequence[Future[Any]] | None = ...,
        **kwargs: Any,
    ) -> Mapping[str, list[str]]: ...
    async def has_what(
        self,
        workers: str | Sequence[str] | None = ...,
    ) -> Mapping[str, list[str]]: ...
    @overload
    async def scatter(
        self,
        data: Mapping[KT, T],
        *,
        broadcast: bool = ...,
        direct: bool = ...,
        hash: bool | None = ...,
    ) -> MutableMapping[KT, Future[T]]: ...
    @overload
    async def scatter(
        self,
        data: Sequence[T],
        *,
        broadcast: bool = ...,
        direct: bool = ...,
        hash: bool | None = ...,
    ) -> list[Future[T]]: ...
    @overload
    async def compute(
        self,
        collection: Delayed[T],
        *,
        optimize_graph: bool = ...,
        scheduler: str | None = ...,
    ) -> T: ...
    @overload
    async def compute(
        self,
        collection: Any,
        *,
        optimize_graph: bool = ...,
        scheduler: str | None = ...,
    ) -> Any: ...
    def persist(
        self,
        collection: PersistableT,
        *,
        optimize_graph: bool = ...,
        scheduler: str | None = ...,
        traverse: bool | None = ...,
    ) -> PersistableT: ...
    def futures_of(self, collection: Any) -> list[Future[Any]]: ...

class LocalCluster:
    """In-process cluster useful for tests and experimentation."""

    asynchronous: bool
    scheduler: Scheduler
    workers: MutableMapping[str, Worker]
    address: str

    def __init__(
        self,
        *,
        asynchronous: bool = ...,
        processes: bool = ...,
        n_workers: int | None = ...,
        threads_per_worker: int | None = ...,
        scheduler_port: int | str | None = ...,
        dashboard_address: str | None = ...,
        silence_logs: bool | str | None = ...,
        scheduler_kwargs: Mapping[str, Any] | None = ...,
        worker_kwargs: Mapping[str, Any] | None = ...,
        **kwargs: Any,
    ) -> None: ...
    async def __aenter__(self) -> LocalCluster: ...
    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None: ...
    def __enter__(self) -> LocalCluster: ...
    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None: ...
    async def start(self, *args: Any, **kwargs: Any) -> LocalCluster: ...
    async def close(self, timeout: float | None = ...) -> None: ...
    def sync(
        self, func: Callable[..., T], *args: Any, asynchronous: bool | None = ..., **kwargs: Any
    ) -> T: ...

@overload
def wait(
    futures: Future[Any],
    *,
    return_when: str = ...,
    timeout: float | None = ...,
) -> Future[Any]: ...
@overload
def wait(
    futures: Sequence[Future[Any]],
    *,
    return_when: str = ...,
    timeout: float | None = ...,
) -> list[Future[Any]]: ...
def as_completed(
    futures: Iterable[Future[Any]],
    *,
    with_results: bool = ...,
) -> AsCompleted: ...
def default_client() -> Client: ...
def get_client() -> Client: ...

__all__ = [
    "AsyncClient",
    "AsCompleted",
    "Client",
    "Future",
    "LocalCluster",
    "Scheduler",
    "Worker",
    "WorkerPlugin",
    "as_completed",
    "default_client",
    "get_client",
    "wait",
]



from __future__ import annotations

from typing import (
    Any,
    Generic,
    Iterable,
    Iterator,
    Mapping,
    Protocol,
    Sequence,
    TypeVar,
    overload,
)

from ..delayed import Delayed

class _DataFrameLike(Protocol):
    def to_dict(self, *args: Any, **kwargs: Any) -> Mapping[str, Any]: ...

class _SeriesLike(Protocol):
    name: Any

    def to_list(self, *args: Any, **kwargs: Any) -> Sequence[Any]: ...

_FrameT_co = TypeVar("_FrameT_co", bound=_DataFrameLike, covariant=True)
_SeriesT_co = TypeVar("_SeriesT_co", bound=_SeriesLike, covariant=True)
_FrameT = TypeVar("_FrameT", bound=_DataFrameLike)
_SeriesT = TypeVar("_SeriesT", bound=_SeriesLike)

class DataFrame(Generic[_FrameT_co], Iterable[Any]):
    """Distributed tabular collection backed by pandas."""

    @property
    def columns(self) -> Any: ...
    def compute(
        self,
        *,
        scheduler: str | None = ...,
        optimize_graph: bool = ...,
        traverse: bool | None = ...,
        **kwargs: Any,
    ) -> _FrameT_co: ...
    def persist(
        self,
        *,
        scheduler: str | None = ...,
        optimize_graph: bool = ...,
        traverse: bool | None = ...,
        **kwargs: Any,
    ) -> DataFrame[_FrameT_co]: ...
    def head(self, n: int = ...) -> _FrameT_co: ...
    def to_delayed(self) -> list[Delayed[_FrameT_co]]: ...
    def __iter__(self) -> Iterator[Any]: ...

class Series(Generic[_SeriesT_co], Iterable[Any]):
    """Distributed one-dimensional labelled data."""

    name: Any

    def compute(
        self,
        *,
        scheduler: str | None = ...,
        optimize_graph: bool = ...,
        traverse: bool | None = ...,
        **kwargs: Any,
    ) -> _SeriesT_co: ...
    def persist(
        self,
        *,
        scheduler: str | None = ...,
        optimize_graph: bool = ...,
        traverse: bool | None = ...,
        **kwargs: Any,
    ) -> Series[_SeriesT_co]: ...
    def to_delayed(self) -> list[Delayed[_SeriesT_co]]: ...
    def __iter__(self) -> Iterator[Any]: ...

@overload
def from_delayed(
    dfs: Sequence[Delayed[_FrameT]] | Delayed[_FrameT],
    *,
    meta: _FrameT | None = ...,
    divisions: Sequence[Any] | None = ...,
    sort: bool = ...,
    verify_meta: bool = ...,
) -> DataFrame[_FrameT]: ...
@overload
def from_delayed(
    series: Sequence[Delayed[_SeriesT]] | Delayed[_SeriesT],
    *,
    meta: _SeriesT,
    divisions: Sequence[Any] | None = ...,
    sort: bool = ...,
    verify_meta: bool = ...,
) -> Series[_SeriesT]: ...

__all__ = ["DataFrame", "Series", "from_delayed"]



from __future__ import annotations

from . import array as array
from . import bag as bag
from . import dataframe as dataframe
from . import distributed as distributed
from .delayed import Delayed, compute, delayed, persist

__all__ = [
    "Delayed",
    "array",
    "bag",
    "compute",
    "dataframe",
    "delayed",
    "distributed",
    "persist",
]



from __future__ import annotations

from typing import Any, Generic, Iterable, Iterator, Sequence, TypeVar

from ..delayed import Delayed

T_co = TypeVar("T_co", covariant=True)

class Bag(Generic[T_co], Iterable[T_co]):
    """Distributed collection of Python objects."""

    def compute(
        self,
        *,
        scheduler: str | None = ...,
        optimize_graph: bool = ...,
        traverse: bool | None = ...,
        **kwargs: Any,
    ) -> list[T_co]: ...
    def persist(
        self,
        *,
        scheduler: str | None = ...,
        optimize_graph: bool = ...,
        traverse: bool | None = ...,
        **kwargs: Any,
    ) -> Bag[T_co]: ...
    def to_delayed(self) -> list[Delayed[list[T_co]]]: ...
    def __iter__(self) -> Iterator[T_co]: ...

def from_delayed(
    bags: Sequence[Delayed[Iterable[T_co]]],
    *,
    length: int | None = ...,
) -> Bag[T_co]: ...

__all__ = ["Bag", "from_delayed"]
